# create database pgconf
# \c pgconf
# psql -U postgres -d pgconf
# python manage.py migrate
# python manage.py shell_plus --print-sql

from polls.models import Choice, Question
from django.utils import timezone

# Simple query, lists nothing as the database is empty
Question.objects.all()

# No hot updates yet
# select n_tup_hot_upd from pg_stat_user_tables where relname = 'polls_question';

# Create the first question
q = Question(question_text="What's new?", pub_date=timezone.now())
# Save will insert, as there is no row to update
q.save()
# No hot updates, as that was an insert
# select n_tup_hot_upd from pg_stat_user_tables where relname = 'polls_question';
# We note the ctid
# select ctid, * from polls_question;

# what if we save again?
q.save()
# ctid changes
# select ctid, * from polls_question;

# the update was hot as the table is not indexed
# select n_tup_hot_upd from pg_stat_user_tables where relname = 'polls_question';

# what happens if we save again?
q.save()


# the update is performed, listing all columns even though we changed nothing in the object.

# you can avoid this by updating just a single field
Question.objects.filter(pk=1).update(question_text='test this')
# select ctid, * from polls_question;
q.save()
# select ctid, * from polls_question;

# a save with update_fields will also only hit a single field
q.save(update_fields=['question_text'])

# select ctid, * from polls_question;

# Let's look deeper
# create extension pageinspect;

# We can list the hidden fields and the tuple contents, including dead rows using the extension
# SELECT t_ctid, t_xmin, t_xmax, tuple_data_split('polls_question'::regclass, t_data, t_infomask, t_infomask2, t_bits) FROM heap_page_items(get_raw_page('polls_question', 0));

# There is only one index on the table
# \d polls_question

# we can list the contents of the index also
# SELECT ctid, data  FROM bt_page_items(get_raw_page('polls_question_pkey', 1));

# Time to create a second one, on updated_at which is an auto updated field in the model
# create index polls_2023 on polls_question(id) where updated_at > '2023-01-01';

# SELECT ctid, data  FROM bt_page_items(get_raw_page('polls_question_pkey', 1));
# SELECT ctid, data  FROM bt_page_items(get_raw_page('polls_2023', 1));
# select n_tup_hot_upd from pg_stat_user_tables where relname = 'polls_question';

# With the index present, run save
q.save()

# notice how it prevents hot updates from happening
# select n_tup_hot_upd from pg_stat_user_tables where relname = 'polls_question';
# SELECT ctid, data  FROM bt_page_items(get_raw_page('polls_question_pkey', 1));
# SELECT ctid, data  FROM bt_page_items(get_raw_page('polls_2023', 1));
# SELECT t_ctid, t_xmin, t_xmax, tuple_data_split('polls_question'::regclass, t_data, t_infomask, t_infomask2, t_bits) FROM heap_page_items(get_raw_page('polls_question', 0));
